import os
import pytest
import httpx
from dotenv import load_dotenv


SERVICE = {
    "api": "merge/",
    "clickhouse": "ping"
}

load_dotenv(".env.test")


def is_responsive(docker_ip: str, port: int, service: str) -> bool:
    try:
        print(f"http://{docker_ip}:{port}/{SERVICE[service]}")
        client = httpx.get(f"http://{docker_ip}:{port}/{SERVICE[service]}")
        if client.status_code == 200:
            return True
        return False
    except httpx.RequestError as exc:
        print(f"An error occurred while requesting {exc.request.url!r}.")


@pytest.fixture(scope="session")
def test_env_vars():
    return {
        "tesseract_backend": os.getenv("TESSERACT_BACKEND", "clickhouse://readonly:test@localhost:9000/default"),
        "tesseract_schema": os.getenv("TESSERACT_SCHEMA", "./tests/fixtures/schema"),
        "clickhouse_user": os.getenv("CLICKHOUSE_USER", "readonly"),
        "clickhouse_password": os.getenv("CLICKHOUSE_PASSWORD", "test")
    }


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return os.path.join(str(pytestconfig.rootdir), "docker-compose.test.yaml")


@pytest.fixture(scope="session")
def clickhouse_service(docker_ip, docker_services):
    service = "clickhouse"
    port = docker_services.port_for(service, 8123)

    docker_services.wait_until_responsive(
        timeout=30.0,
        pause=0.5,
        check=lambda: is_responsive(docker_ip, port, service)
    )

    return docker_ip, port

@pytest.fixture(scope="session")
def api_service(docker_ip, docker_services):
    service = "api"
    port = docker_services.port_for(service, 8000)

    docker_services.wait_until_responsive(
        timeout=30.0,
        pause=0.5,
        check=lambda: is_responsive(docker_ip, port, service)
    )

    return docker_ip, port


from pathlib import Path

import tomli


def read_version() -> str:
    # Load the version from pyproject.toml
    with Path("pyproject.toml").open("rb") as fileio:
        pyproject = tomli.load(fileio)
    return pyproject["project"]["version"]


def write_in_library() -> None:
    version = read_version()

    file_path = Path("src/logiclayer_merge", "__init__.py")

    with file_path.open() as f:
        lines = f.readlines()

    with file_path.open("w") as f:
        for line in lines:
            if line.startswith("__version__"):
                f.write(f'__version__ = "{version}"\n')
            else:
                f.write(line)

    print(f"Version updated in __init__.py to {version}")


if __name__ == "__main__":
    write_in_library()


__title__ = "logiclayer-merge"
__description__ = "Allows to merge the results of other LogicLayer modules in a single response."
__version__ = "0.1.0"

__all__ = ("MergeModule", "ResponseFormat")

from .module import MergeModule



from typing import Annotated, Optional

import datetime
import logiclayer as ll
import polars as pl
import httpx
from urllib.parse import urlparse

from fastapi import Depends, Header, Query, status
from fastapi.responses import JSONResponse
from tesseract_olap import OlapServer
from tesseract_olap.logiclayer.response import data_response, ResponseFormat
from tesseract_olap.schema.public import TesseractSchema, TesseractCube
from tesseract_olap.schema import DataType

from src.logiclayer_merge import __title__, __version__
from src.logiclayer_merge.models import JoinStep, MultiRequest, MergeParams, CubeRequest, ColumnsRequest, Result, FileParams
from src.logiclayer_merge.exceptions import DomainNotAllowedException, InvalidUrlException


def auth_token(
    header_auth: Annotated[Optional[str], Header(alias="authorization")] = None,
    header_jwt: Annotated[Optional[str], Header(alias="x-tesseract-jwt")] = None,
    query_token: Annotated[Optional[str], Query(alias="token")] = None,
):
    if header_jwt:
        return ll.AuthToken(ll.AuthTokenType.JWTOKEN, header_jwt)
    if query_token:
        return ll.AuthToken(ll.AuthTokenType.SEARCHPARAM, query_token)
    if header_auth:
        if header_auth.startswith("Bearer "):
            return ll.AuthToken(ll.AuthTokenType.JWTOKEN, header_auth[7:])
        if header_auth.startswith("Basic "):
            return ll.AuthToken(ll.AuthTokenType.BASIC, header_auth[6:])
        if header_auth.startswith("Digest "):
            return ll.AuthToken(ll.AuthTokenType.DIGEST, header_auth[7:])
    return None


class MergeModule(ll.LogicLayerModule):
    """Endpoint merging module.

    Allows to merge the results of other LogicLayer modules in a single response.
    """

    def __init__(self, olap: Optional[OlapServer] = None, allowed_domains: Optional[str] = "", **kwargs):
        super().__init__(**kwargs)
        self.tesseract = olap
        self.cubes = None
        self.adjacency_list = None
        self.ad_cube_measure = None
        self.cube_to_measures = None
        self.allowed_domains = allowed_domains

    def startup_tasks(self):
        """Startup tasks."""
        self.set_allowed_domains()
        self.set_cubes()
        self.build_adjacency_list()
        self.preprocess_cube()

    def set_allowed_domains(self):
        """Set the allowed domains for this module."""
        if self.allowed_domains != "":
            self.allowed_domains = self.allowed_domains.split(',')

    def set_cubes(self, cubes: list[str] = None):
        """Set the cube registry for this module."""
        if cubes is None:
            cubes = None
        self.cubes = TesseractSchema.from_entity(self.tesseract.schema, show_all=True).cubes

    def build_adjacency_list(self):
        adjacency_list = {}
        for cube in self.cubes:
            cube_name = cube.name
            adjacency_list[cube_name] = []
            levels = [
                level 
                for dim in cube.dimensions 
                for hier in dim.hierarchies
                for level in hier.levels]
            for lvl in levels:
                level_name = lvl.name
                adjacency_list[cube_name].append(level_name)
        reversed_adj = {}
        for cube, lvl in adjacency_list.items():
            for level in lvl:
                reversed_adj[level] = reversed_adj.get(level, []) + [cube]

        self.adjacency_list =  {**adjacency_list, **reversed_adj}

    def preprocess_cube(self):
        ad_cube_measure = {}
        cube_to_measures = {}

        for c in self.cubes:
            for m in c.measures:
                ad_cube_measure[c.name] = ad_cube_measure.get(c.name, []) +[m.name+ "@"+c.name]
                cube_to_measures[c.name] = cube_to_measures.get(c.name, []) +[m.name]
        
        self.ad_cube_measure = ad_cube_measure
        self.cube_to_measures = cube_to_measures

    @ll.route("GET", "/")
    def status(self) -> ll.ModuleStatus:
        """Return the current status of the internals of this module."""
        return ll.ModuleStatus(module=__title__, version=__version__, debug=self.debug, status="ok")

    @ll.route("POST", "/query.{extension}")
    def multiquery_data(
        self,
        extension: ResponseFormat,
        params: MultiRequest,
        token: Optional[ll.AuthToken] = Depends(auth_token),
    ):
        """Execute a request for joining data from the server."""
        # Add role data to all params, overwrite possible user injections
        roles = self.auth.get_roles(token)
        for request in params.requests:
            request.roles = roles

        query = params.build_query(self.tesseract.schema)

        with self.tesseract.session() as session:
            result = session.fetch_dataframe(query.initial)
            step = JoinStep.new(result)

            for query_right, join_params in query.join_with:
                result = session.fetch_dataframe(query_right)
                join_params.suffix = f"_{query_right.cube.name}"
                step = step.join_with(result, join_params)

        result = step.get_result(params.pagination)

        return data_response(
            query,
            result,
            extension
        )

    @ll.route("GET", "/structures")
    def get_data_structures(
        self, 
    ):
        """Send adjacency list and cube to measures mapping"""
        return JSONResponse(content=
            {
                "ad_cube_measure": self.ad_cube_measure,
                "cube_to_measures": self.cube_to_measures,
                "adjacency_list": self.adjacency_list,
            },
        )

    @ll.route("GET", "/cubes")
    def get_cubes(
        self,
    ):
        """send available cubes"""
        return self.cubes
    
    @ll.route("GET", "/allowed")
    def get_allowed_domains(
        self,
    ):
        """send allowed domains"""
        return self.allowed_domains

    @ll.route("POST", "/cubes/dimensions_measures")
    def get_cubes_dim(
        self,
        r: CubeRequest
    ):
        """For selected cubes send dimensions and measures"""
        try:
            data = []
            for cube in r.cube_name:
                cube = self.tesseract.schema.get_cube(cube)
                cube= TesseractCube.from_entity(cube, locale=r.locale, show_all=r.show_all)
                data.append(cube)
            return data

        except:
            return JSONResponse({"Status": "ok", "Message": "Cube not found"})

    @ll.route("POST", "/cubes/merge_options")
    def get_merge_options(
        self,
        r: ColumnsRequest
    ):
        """for dimensions selected send available cubes to merge on or available cubes+measures"""
        try:
            option_sets = list(set.intersection(*[set(self.adjacency_list.get(c, [])) for c in r.columns]))
            if option_sets:
                measures_lists = [self.cube_to_measures.get(c, []) for c in option_sets]
                option_measures = list(set.intersection(*[set(measures) for measures in measures_lists]))
            else:
                option_measures = []
            return JSONResponse({"option_sets": option_sets, "option_measures": option_measures})

        except:
            return JSONResponse({"Status": "ok", "Message": "No Match Found"})

    @ll.route("POST", "/cubes/merge.{extension}")
    async def merge_cubes(
        self,
        extension: ResponseFormat,
        merge_params: MergeParams
    ):
        try:
            join_params = merge_params.join
            left_params = merge_params.query_left
            right_params = merge_params.query_right
            pagi = merge_params.pagination
            df, annotations = await self.fetch_query(left_params.url, left_params.data, left_params.headers)
            df2, annotations2 = await self.fetch_query(right_params.url, right_params.data, right_params.headers)

            date = datetime.datetime.now(tz=datetime.timezone.utc)

            query = FileParams(
                filename = f"datajoin_{date.strftime(r'%Y-%m-%d_%H-%M-%S')}",
                annotations = {"query_left": annotations, "query_right": annotations2}
            )

            step = JoinStep(data=df, keys=[], statuses=[])

            result = Result(
                data=df2, 
                columns={
                    k: DataType.from_polars(v) for k, v in dict(zip(df.columns, df.dtypes)).items()
                },
                cache={"key": "", "status": ""}
            )

            step = step.join_with(result, join_params)

            result = step.get_result(pagi)

            return data_response(
                query,
                result,
                extension
            )

        except Exception as e:
            return JSONResponse(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                                content={"Query left": str(merge_params.query_left),
                                 "Query right": str(merge_params.query_right),
                                 "Error": str(e)})

    async def fetch_query(self, url, data, headers):

        self._validate_url(url)

        async with httpx.AsyncClient() as client:
            r = await client.get(url=url, auth=data, headers=headers)
            content_type = r.headers.get("content-type", "")

            if "text/csv" in content_type or content_type.endswith("/csv"):
                return pl.read_csv(r.content, separator=",", has_header=True, infer_schema_length=1000), "Not Available"
            elif "application/json" in content_type:
                json_data = r.json()
                return pl.DataFrame(json_data["data"]), json_data["annotations"]
            else:
                return "Format not Supported"

    def _validate_url(self, url: str) -> None:
        """Validate URL and raise appropriate exceptions"""
        try:
            parsed = urlparse(url)
            if not parsed.scheme or not parsed.netloc:
                raise InvalidUrlException(url, "Missing scheme or network location")
        except Exception:
            raise InvalidUrlException(url, "Malformed URL")

        # Check domain
        domain = parsed.netloc.split(':')[0]
        valid_domains = [d for d in self.allowed_domains if d and isinstance(d, str)]

        if not any(
            domain.endswith(f".{allowed}") or domain == allowed
            for allowed in valid_domains
        ):
            raise DomainNotAllowedException(domain, valid_domains, source=url)


from fastapi import status
from fastapi.responses import JSONResponse
from typing import Any, Dict, Optional


class APIException(Exception):
    """Base exception that can return JSON response directly"""
    def __init__(
        self,
        status_code: int = 500,
        error_type: str = "api_error",
        message: str = "An error occurred",
        details: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None
    ):
        self.status_code = status_code
        self.error_type = error_type
        self.message = message
        self.details = details or {}
        self.headers = headers or {}
        super().__init__(self.message)

    def to_json_response(self) -> JSONResponse:
        """Convert exception to FastAPI JSONResponse"""
        content = {
            "error": self.error_type,
            "message": self.message,
            **self.details
        }
        return JSONResponse(
            status_code=self.status_code,
            content=content,
            headers=self.headers
        )

class DomainNotAllowedException(APIException):
    def __init__(self, domain: str, allowed_domains: list, source: Optional[str] = None):
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            error_type="invalid_domain",
            message=f"Domain '{domain}' is not allowed",
            details={
                "domain": domain,
                "allowed_domains": allowed_domains,
                "source": source,
                "suggestion": f"Please use one of the allowed domains: {', '.join(allowed_domains)}"
            }
        )

class InvalidUrlException(APIException):
    def __init__(self, url: str, reason: str = "Invalid format"):
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            error_type="invalid_url",
            message=f"Invalid URL: {url}",
            details={
                "url": url,
                "reason": reason,
                "suggestion": "Please provide a valid URL with protocol (http:// or https://)"
            }
        )

class MergeException(Exception):
    """Base class for exceptions in this module."""



import httpx

class TestAPIClient:
    def test_ping_api_successful(self, api_service):
        docker_ip, port = api_service
        client = httpx.get(f"http://{docker_ip}:{port}/merge/")

        assert client.status_code == 200

    def test_endpoint_cubes_fastapi_successful(self, api_service):
        docker_ip, port = api_service
        client = httpx.get(f"http://{docker_ip}:{port}/merge/cubes")

        assert client.status_code == 200

    def test_endpoint_structures_fastapi_successful(self, api_service):
        docker_ip, port = api_service
        client = httpx.get(f"http://{docker_ip}:{port}/merge/structures")

        assert client.status_code == 200

    def test_endpoint_dimensions_measures_fastapi_successful(self, api_service):
        docker_ip, port = api_service
        data = '{"cube_name":["customers","orders"],"locale": "en"}'
        client = httpx.post(f"http://{docker_ip}:{port}/merge/cubes/dimensions_measures", content=data)

        assert client.status_code == 200

    def test_endpoint_merge_fastapi_error(self, api_service):
        docker_ip, port = api_service
        data = """{
            "query_left": {
                "url": "https://api-v2.oec.world/tesseract/data.csv?locale=en&cube=trade_i_baci_a_96&measures=Trade+Value&drilldowns=Year&time=Year.latest.10"
            },
            "query_right": {
                "url": "https://api-v2.oec.world/tesseract/data.jsonrecords?locale=en&cube=trade_i_baci_a_22&measures=Trade+Value&drilldowns=Year&time=Year.latest.10"
            },
            "pagination": {
                "limit": 0,
                "offset": 0
            },
            "join": {
                "on": "Year",
                "how": "left",
                "suffix": " (right)"
            }
        }"""
        client = httpx.post(f"http://{docker_ip}:{port}/merge/cubes/merge.jsonrecords", content=data)

        assert client.status_code == 500



__title__ = "logiclayer-merge"
__description__ = "Logiclayer-Merge instance"

import os
import logging

from logiclayer import LogicLayer
from src.logiclayer_merge import MergeModule
from tesseract_olap import OlapServer

logging.basicConfig(level=logging.DEBUG)

olap_backend = os.environ["TESSERACT_BACKEND"]
olap_schema = os.environ["TESSERACT_SCHEMA"]
allowed_domains = os.environ.get("ALLOWED_DOMAINS", "")

olap = OlapServer(backend=olap_backend, schema=olap_schema)
mod = MergeModule(olap=olap, allowed_domains=allowed_domains)
mod.startup_tasks()
layer = LogicLayer()
layer.add_module('/merge', mod)



import json
import httpx

class TestClickhouseClient:
    def test_ping_clickhouse_successful(self, clickhouse_service):
        docker_ip, port = clickhouse_service
        client = httpx.get(f"http://{docker_ip}:{port}/ping")

        assert client.status_code == 200

    def test_table_orders(self, clickhouse_service, test_env_vars):
        clickhouse_user = test_env_vars.get("clickhouse_user")
        clickhouse_password = test_env_vars.get("clickhouse_password")
        docker_ip, port = clickhouse_service
        payload = 'select * from orders;'
        client = httpx.post(f'http://{docker_ip}:{port}/?add_http_cors_header=1&user={clickhouse_user}&password={clickhouse_password}&default_format=JSON&enable_http_compression=1&extremes=1',
                            content=payload)

        assert len(json.loads(client.content.decode('utf-8'))["data"]) == 9

    def test_table_customers(self, clickhouse_service, test_env_vars):
        clickhouse_user = test_env_vars.get("clickhouse_user")
        clickhouse_password = test_env_vars.get("clickhouse_password")
        docker_ip, port = clickhouse_service
        payload = 'select * from customers;'
        client = httpx.post(f'http://{docker_ip}:{port}/?add_http_cors_header=1&user={clickhouse_user}&password={clickhouse_password}&default_format=JSON&enable_http_compression=1&extremes=1',
                            content=payload)

        assert len(json.loads(client.content.decode('utf-8'))["data"]) == 6
    
    def test_table_join(self, clickhouse_service, test_env_vars):
        clickhouse_user = test_env_vars.get("clickhouse_user")
        clickhouse_password = test_env_vars.get("clickhouse_password")
        docker_ip, port = clickhouse_service
        payload = 'select * from customers left outer join orders on customers.customer_id = orders.customer_id;'
        client = httpx.post(f'http://{docker_ip}:{port}/?add_http_cors_header=1&user={clickhouse_user}&password={clickhouse_password}&default_format=JSON&enable_http_compression=1&extremes=1',
                            content=payload)

        assert len(json.loads(client.content.decode('utf-8'))["data"]) == 10



from collections.abc import Sequence
from typing import Literal, Optional, TypedDict, Union

import hashlib
import polars as pl
from pydantic import Field, model_validator
from pydantic import BaseModel as PydanticBaseModel
from tesseract_olap.query import DataMultiQuery, DataRequest
from tesseract_olap.schema import SchemaTraverser, DataType
from tesseract_olap.backend.models import Result


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True


JoinStrategy = Literal[
    "inner",
    "left",
    "right",
    "full",
    "semi",
    "anti",
    "cross",
    "outer",
]


class JoinOnColumns(BaseModel):
    """Specifies the columns to be used in a merge operation, when their names are different."""

    left_on: Union[str, list[str]]
    right_on: Union[str, list[str]]


class JoinIntent(BaseModel):
    """Specifies the intent of the user to perform a Join operation between 2 datasets."""

    on: Union[str, list[str], JoinOnColumns, None] = None
    how: JoinStrategy = "left"
    suffix: Optional[str] = None
    validate_relation: Literal["m:m", "m:1", "1:m", "1:1"] = "m:m"
    join_nulls: bool = False
    coalesce: Optional[bool] = None


class PaginationIntent(BaseModel):
    limit: int = 0
    offset: int = 0


class MultiRequest(BaseModel):
    """Describes a request done by the user."""

    requests: list[DataRequest]
    joins: list[JoinIntent] = Field(default_factory=list)
    pagination: PaginationIntent = Field(default_factory=PaginationIntent)

    @model_validator(mode="before")
    @classmethod
    def parse_request(cls, value: object):
        if isinstance(value, dict):
            requests = value.get("queries") or value.get("requests", [])
            if not isinstance(requests, Sequence):
                msg = "Invalid 'requests' parameter: it must be a list of dictionaries containing the request parameters of the queries to be merged."
                raise ValueError(msg)

            request_count = len(requests)
            if not request_count > 1:
                msg = "At least 2 DataRequest objects are required to perform a join operation."
                raise ValueError(msg)

            joins = value.get("joins", [])
            if not isinstance(joins, Sequence):
                msg = "Invalid 'joins' parameter. It must be a list of dictionaries with the parameters to use."
                raise ValueError(msg)

            if not joins:
                joins = [{}] * (request_count - 1)
            elif len(joins) == 1:
                joins = list(joins) * (request_count - 1)
            elif len(joins) == request_count - 1:
                pass
            else:
                msg = f"Invalid 'joins' parameter. It must be a list of objects with the parameters to use; this list must contain 1 object (if you intend to apply the same parameters to all queries), {request_count - 1} objects (one per each step of this join operation), or left empty/unset to let the server attempt to guess the parameters."
                raise ValueError(msg)

            return {
                "requests": requests,
                "joins": joins,
                "pagination": value.get("pagination", "0,0"),
            }
        return value

    def build_query(self, schema: SchemaTraverser) -> "DataMultiQuery":
        """Generate a DataMultiQuery object from the params in this request."""
        return DataMultiQuery.from_requests(schema, self.requests, self.joins)


class _JoinParameters(TypedDict, total=False):
    """Describe the keyword args needed for a polars.DataFrame.join operation."""

    on: Union[str, list[str]]
    coalesce: Optional[bool]
    join_nulls: bool
    left_on: Union[str, list[str]]
    right_on: Union[str, list[str]]
    suffix: str
    validate: Literal["m:m", "m:1", "1:m", "1:1"]


class JoinStep:
    data: pl.DataFrame
    keys: list[str]
    statuses: list[str]

    def __init__(
        self,
        data: pl.DataFrame,
        *,
        keys: list[str],
        statuses: list[str],
    ):
        self.data = data
        self.keys = keys
        self.statuses = statuses

    def join_with(self, result: Result, join: JoinIntent):
        params: _JoinParameters = {
            "suffix": join.suffix or "_",
            "validate": join.validate_relation,
            "join_nulls": join.join_nulls,
            "coalesce": join.coalesce,
        }

        if isinstance(join.on, (str, list)):
            params.update(on=join.on)
        elif isinstance(join.on, JoinOnColumns):
            params.update(left_on=join.on.left_on, right_on=join.on.right_on)

        return JoinStep(
            self.data.join(result.data, how=join.how, **params),
            keys=[*self.keys, result.cache["key"]],
            statuses=[*self.statuses, result.cache["status"]],
        )

    def get_result(self, pagi: PaginationIntent):
        df = self.data

        cache_key = "/".join(self.keys).encode("utf-8")
        return Result(
            data=df.slice(pagi.offset, pagi.limit or None),
            columns={
                k: DataType.from_polars(v) for k, v in dict(zip(df.columns, df.dtypes)).items()
            },
            cache={
                "key": hashlib.md5(cache_key, usedforsecurity=False).hexdigest(),
                "status": ",".join(self.statuses),
            },
            page={"limit": pagi.limit, "offset": pagi.offset, "total": df.height},
        )

    @classmethod
    def new(cls, result: Result):
        return cls(
            result.data,
            keys=[result.cache["key"]],
            statuses=[result.cache["status"]],
        )


class QueryRequest(BaseModel):
    url: str
    data: Optional[str] = None
    headers: Optional[str] = None


class MergeParams(BaseModel):
    query_left : QueryRequest
    query_right: QueryRequest
    pagination: PaginationIntent = Field(default_factory=PaginationIntent)
    join: JoinIntent


class CubeRequest(BaseModel):
    cube_name: list[str]
    locale: str
    show_all: Optional[bool] = False


class ColumnsRequest(BaseModel):
    columns: list[str]


class FileParams(BaseModel):
    filename: str
    annotations: dict

    def get_annotations(self):
        return self.annotations